// ****************************************************************************
//  Project:        GUYMAGER
// ****************************************************************************
//  Programmer:     Guy Voncken
//                  Police Grand-Ducale
//                  Service de Police Judiciaire
//                  Section Nouvelles Technologies
// ****************************************************************************

#ifndef __DEVICE_H__
#define __DEVICE_H__

#include <openssl/md5.h> //lint !e537 Repeated include

#include <parted/parted.h>
#include <QMetaType>
#include <QReadWriteLock>
#include <QMutex>
#include <QTime>

#include "fifo.h"
#include "info.h"
#include "file.h"

class t_ThreadRead;
class t_ThreadMD5;
class t_ThreadCompress;
class t_ThreadWrite;

#define DEVICE_DEFAULT_SECTOR_SIZE ((size_t)512)

class t_Device
{
   public:
      typedef enum
      {
         Connected,
         Disconnected,
      } t_DeviceState;

      typedef enum
      {
         Idle,
         Running,
         Paused,
         Finished,
         Cleanup,
         Aborted
      } t_AcquireState;

      typedef enum
      {
         None,
         UserRequest,
         ThreadWriteFileError,
         ThreadReadFileError
      } t_AbortReason;

      class t_Acquisition
      {
         public:
            QString         Path;             // Always end with /
            QString         ImageFilename;    // Image filename, without extension
            QString         InfoFilename;     // Info filename, without extension
            t_File::Format  Format;
            bool            CalcMD5;
            bool            VerifyMD5;
            QString         CaseNumber;
            QString         EvidenceNumber;
            QString         Examiner;
            QString         Description;
            QString         Notes;
      };

   public:
      t_Device (const QString &SerialNumber=QString(), const PedDevice *pPedDev=NULL);
      t_Device (const QString &SerialNumber, const QString &LinuxDevice, const QString &Model,
                                             quint64 SectorSize, quint64 SectorSizePhys, quint64 Size=0);
     ~t_Device ();

      static QVariant GetSerialNumber  (t_pDevice pDevice);
      static QVariant GetLinuxDevice   (t_pDevice pDevice);
      static QVariant GetModel         (t_pDevice pDevice);
      static QVariant GetDeviceState   (t_pDevice pDevice);
      static QVariant GetAcquireState  (t_pDevice pDevice);
      static QVariant GetSectorSize    (t_pDevice pDevice);
      static QVariant GetSectorSizePhys(t_pDevice pDevice);
      static QVariant GetBadSectorCount(t_pDevice pDevice);
      static QVariant GetSize          (t_pDevice pDevice);
      static QVariant GetSizeHuman     (t_pDevice pDevice);
      static QVariant GetSizeHumanFrac (t_pDevice pDevice, int FracLen=0);
      static QVariant GetProgress      (t_pDevice pDevice);
      static QVariant GetCurrentSpeed  (t_pDevice pDevice);
      static QVariant GetAverageSpeed  (t_pDevice pDevice);
      static QVariant GetRemaining     (t_pDevice pDevice);
      static QVariant GetFifoStatus    (t_pDevice pDevice);

      inline void SetCurrentWritePos (quint64 Pos)
      {
         SemCurrentWritePos.lockForWrite ();
         CurrentWritePos = Pos;
         SemCurrentWritePos.unlock ();
      }

      inline void IncCurrentWritePos (int Inc)
      {
         SemCurrentWritePos.lockForWrite ();
         CurrentWritePos += Inc; //lint !e737 Loss of sign
         SemCurrentWritePos.unlock ();
      }

      inline quint64 GetCurrentWritePos (void)
      {
         quint64 Pos;
         SemCurrentWritePos.lockForRead ();
         Pos = CurrentWritePos;
         SemCurrentWritePos.unlock ();
         return Pos;
      }

      inline void AddBadSector (quint64 Sector)
      {
         SemBadSectors.lock ();
         BadSectors.append(Sector);
         SemBadSectors.unlock ();
      }

      APIRET GetBadSectors (QList<quint64> &BadSectorsCopy)
      {
         SemBadSectors.lock ();
         BadSectorsCopy = BadSectors;
         SemBadSectors.unlock ();
         return NO_ERROR;
      }

      quint64 GetBadSectorCount (void)
      {
         quint64 Count;
         SemBadSectors.lock ();
         Count = BadSectors.count();  //lint !e732 Loss of sign
         SemBadSectors.unlock ();
         return Count;
      }

      APIRET ClearBadSectors (void)
      {
         SemBadSectors.lock ();
         BadSectors.clear();
         SemBadSectors.unlock ();
         return NO_ERROR;
      }

      APIRET GetMessage (QString &MessageCopy)
      {
         SemMessage.lock ();
         MessageCopy = Message;
         SemMessage.unlock ();
         return NO_ERROR;
      }

      APIRET SetMessage (const QString &NewMessage)
      {
         SemMessage.lock ();
         Message = NewMessage;
         SemMessage.unlock ();
         return NO_ERROR;
      }

      bool HasHashThread         (void) const;
      bool HasCompressionThreads (void) const;

   private:
      void Initialise (void);
      void InitialiseDeviceSpecific (const QString &SerialNumber, const QString &LinuxDevice, const QString &Model,
                                    quint64 SectorSize, quint64 SectorSizePhys, quint64 Size);

   public:  QString                   SerialNumber;
            QString                   Model;
            QString                   LinuxDevice;    // for instance /dev/hda
            bool                      Local;
            quint64                   SectorSize;
            quint64                   SectorSizePhys;
            quint64                   Size;
            bool                      Removable;

            t_DeviceState             DeviceState;
            t_AcquireState            AcquireState;
   private: QString                   Message;          // This field can be used by the threads to communicate messages displayed in spreadsheet field "acquisition state"
   public:  bool                      AbortRequest;
            t_AbortReason             AbortReason;
            bool                      DeleteAfterAbort;

            t_Acquisition             Acquisition;
            t_Info                    Info;             // Must only be used if Aqcuisition.InfoFilename is valid!

            t_ThreadRead            *pThreadRead;
            t_pFifo                  pFifoRead;         // The Fifo that is used by ThreadRead
            FILE                    *pFileSrc;
            quint64                   CurrentReadPos;
   private: QList<quint64>            BadSectors;       // Accessed concurrently, use appropriate functions
   public:  bool                      FallbackMode;     // Set if an error occurs while reading a large block. If set, sectors are read individually until the next large block boundary.
            QDateTime                 StartTimestamp;

            t_ThreadMD5             *pThreadMD5;
            t_pFifo                  pFifoMD5In;        // The Fifo that is used by ThreadMD5
            t_pFifo                  pFifoMD5Out;       // The Fifo that is used by ThreadMD5
            unsigned char             MD5Digest[MD5_DIGEST_LENGTH];

            QList<t_ThreadCompress *> ThreadCompressList;
            t_pFifoCompressIn        pFifoCompressIn;
            t_pFifoCompressOut       pFifoCompressOut;

            t_ThreadWrite           *pThreadWrite;
   private: quint64                   CurrentWritePos;  // Accessed concurrently, use appropriate function
   public:  t_pFifo                  pFifoWrite;        // The Fifo that is used by ThreadWrite. This is the same then pFifoRead if not MD5 and E01 compression is calculated.
//            FILE                    *pFileDst;
            QDateTime                 StopTimestamp;

            quint64                   PrevWritePos;
            QTime                     PrevTimestamp;    // Used for calculating current speed
            double                    PrevSpeed;        // Used for calculating current speed
            bool                      Checked;          // Helper variable for matching algorithm in SlotScanFinished, not used elsewhere.

   private: QReadWriteLock            SemCurrentWritePos;
            QMutex                    SemBadSectors;
            QMutex                    SemMessage;
};

class t_DeviceList: public QList<t_pDevice>
{
   public:
      t_DeviceList (void);
     ~t_DeviceList ();

      t_pDevice AppendNew (const QString &SerialNumber, const PedDevice *pPedDev);
      t_pDevice AppendNew (const QString &SerialNumber, const QString &LinuxDevice, const QString &Model,
                           quint64 SectorSize, quint64 SectorSizePhys, quint64 Size=0);
      APIRET MatchDevice (t_pcDevice pDevCmp, t_pDevice &pDeviceMatch);
};

typedef t_DeviceList *t_pDeviceList;

Q_DECLARE_METATYPE(t_pDeviceList); //lint !e19 Useless declaration


// ------------------------------------
//             Error codes
// ------------------------------------

enum
{
   ERROR_DEVICE_SERNR_MATCH_MODEL_MISMATCH = ERROR_BASE_DEVICE + 1,
   ERROR_DEVICE_SERNR_MATCH_LENGTH_MISMATCH,
   ERROR_DEVICE_BAD_DEVICE_STATE,
   ERROR_DEVICE_BAD_ACQUIRE_STATE,
   ERROR_DEVICE_BAD_ABORTREASON,
   ERROR_DEVICE_NOT_CLEAN
};


#endif

