// ****************************************************************************
//  Project:        GUYMAGER
// ****************************************************************************
//  Programmer:     Guy Voncken
//                  Police Grand-Ducale
//                  Service de Police Judiciaire
//                  Section Nouvelles Technologies
// ****************************************************************************
//  Module:         Acquisition dialog
// ****************************************************************************

#include <QtGui>

#include "common.h"
#include "compileinfo.h"
#include "config.h"
#include "qtutil.h"
#include "dlgdirsel.h"
#include "dlgacquire.h"
#include "dlgacquire_private.h"

// -----------------------------
//           Constants
// -----------------------------

const char *DLG_ACQUIRE_DEFAULT_EMPTY_FILENAME = "Out";  // Default image file name if - after removing special chars - the file name is empty.

// -----------------------------
//           Classes
// -----------------------------

class t_DlgAcquireLocal
{
   public:
      QRadioButton    *pRadioButtonFormatEWF;
      QRadioButton    *pRadioButtonFormatDD;

      QList<QWidget*>   EwfWidgetsList;   // Used zo comfortably enable / disable EWF related entry fields and labels

      t_pDlgAcquireLineEdit pLineEditDestDirectory;      // used very often, so a pointer to these instaead fo searching in the list each time
      t_pDlgAcquireLineEdit pLineEditDestImageFilename;
      t_pDlgAcquireLineEdit pLineEditDestInfoFilename;

      QCheckBox       *pCheckBoxCalcMD5;
      QCheckBox       *pCheckBoxVerifyMD5;

      QPushButton     *pButtonOk;
      QPushButton     *pButtonCancel;

      QFileDialog     *pDlg;

      t_pDevice        pDevice;
};


// -----------------------------
//    Field utility functions
// -----------------------------

static t_pCfgDlgAcquireField DlgAcquireGetField (const QString &Name)
{
   t_pCfgDlgAcquireFields pDlgAcquireFields;
   t_pCfgDlgAcquireField  pDlgAcquireField;
   int i;

   CHK_EXIT (CfgGetDlgAcquireFields (&pDlgAcquireFields))

   for (i=0; i<pDlgAcquireFields->count(); i++)
   {
      pDlgAcquireField = pDlgAcquireFields->at(i);
      if (pDlgAcquireField->FieldName.compare(Name, Qt::CaseInsensitive)==0)
         return pDlgAcquireField;
   }
   LOG_ERROR ("DlgAcquire field %s not found", QSTR_TO_PSZ(Name))
   return NULL;
}


static t_pDlgAcquireLineEdit DlgAcquireGetLineEdit (const QString &Name)
{
   return DlgAcquireGetField(Name)->pLineEdit;
}


static APIRET DlgAcquireResolveSpecialSequences (t_pDevice pDevice, bool ConsiderFields, const QString &In, QString &Out)
{
   t_pCfgDlgAcquireFields pDlgAcquireFields;
   t_pCfgDlgAcquireField  pDlgAcquireField;
   QDateTime               Now = QDateTime::currentDateTime();
   int                     i;
   QString                 Search;

   // Replace date/time values
   // ------------------------
   Out = In;
   Out.replace ("%d%"   , Now.toString("d"   ));  // Special sequences
   Out.replace ("%dd%"  , Now.toString("dd"  ));
   Out.replace ("%ddd%" , Now.toString("ddd" ));
   Out.replace ("%dddd%", Now.toString("dddd"));
   Out.replace ("%M%"   , Now.toString("M"   ));
   Out.replace ("%MM%"  , Now.toString("MM"  ));
   Out.replace ("%MMM%" , Now.toString("MMM" ));
   Out.replace ("%MMMM%", Now.toString("MMMM"));
   Out.replace ("%yy%"  , Now.toString("yy"  ));
   Out.replace ("%yyyy%", Now.toString("yyyy"));
   Out.replace ("%h%"   , Now.toString("h"   ));
   Out.replace ("%hh%"  , Now.toString("hh"  ));
   Out.replace ("%m%"   , Now.toString("m"   ));
   Out.replace ("%mm%"  , Now.toString("mm"  ));
   Out.replace ("%s%"   , Now.toString("s"   ));
   Out.replace ("%ss%"  , Now.toString("ss"  ));
   Out.replace ("%AP%"  , Now.toString("AP"  ));
   Out.replace ("%ap%"  , Now.toString("ap"  ));

   // Replace device values
   // ---------------------
   QVariant SizeHuman = t_Device::GetSizeHuman (pDevice);
   Out.replace ("%serial%" , pDevice->SerialNumber);
   Out.replace ("%model%"  , pDevice->Model       );
   Out.replace ("%size%"   , SizeHuman.toString() );
   Out.replace ("%version%", pCompileInfoVersion  );

   // Replace field values
   // --------------------
   if (ConsiderFields)
   {
      CHK_EXIT (CfgGetDlgAcquireFields (&pDlgAcquireFields))
      for (i=0; i<pDlgAcquireFields->count(); i++)
      {
         pDlgAcquireField = pDlgAcquireFields->at(i);
         Search = "%" + pDlgAcquireField->FieldName + "%";
         Out.replace (Search, pDlgAcquireField->pLineEdit->text());
      }
   }

   return NO_ERROR;
}


static APIRET DlgAcquireGetFieldValue (t_pDevice pDevice, t_pCfgDlgAcquireField pDlgAcquireField, QString &Set)
{
   if (( pDlgAcquireField->EntryMode == CFG_ENTRYMODE_SHOWLAST) &&
       (!pDlgAcquireField->LastEnteredValue.isNull()))
   {
      Set = pDlgAcquireField->LastEnteredValue;
   }
   else
   {
      Set = pDlgAcquireField->DefaultValue;
      CHK (DlgAcquireResolveSpecialSequences (pDevice, false, pDlgAcquireField->DefaultValue, Set))
   }

   return NO_ERROR;
}

// -----------------------------
//         t_DlgAcquire
// -----------------------------

t_DlgAcquire::t_DlgAcquire ()
{
   CHK_EXIT (ERROR_DLGACQUIRE_CONSTRUCTOR_NOT_SUPPORTED)
} //lint !e1401 pOwn not initialised

t_DlgAcquire::t_DlgAcquire (t_pDevice pDevice, QWidget *pParent, Qt::WFlags Flags)
   :QDialog (pParent, Flags)
{
   static bool             Initialised = false;
   t_pCfgDlgAcquireFields pDlgAcquireFields;
   t_pCfgDlgAcquireField  pDlgAcquireField;
   t_pDlgAcquireLineEdit  pLineEdit     = NULL;
   QPushButton           *pButtonBrowse = NULL;
   QVBoxLayout           *pLayout;
   QLabel                *pLabel;
   QString                 DefaultFilename;
   QString                 Path;
   QString                 Str;
   QString                 ButtonTextEWF;
   QString                 ButtonTextDD;
   QString                 Set;
   int                     Row = 0;
   int                     i;
   bool                    DestDirectory;

   if (!Initialised)
   {
      Initialised = true;
      CHK_EXIT (TOOL_ERROR_REGISTER_CODE (ERROR_DLGACQUIRE_CONSTRUCTOR_NOT_SUPPORTED))
      CHK_EXIT (TOOL_ERROR_REGISTER_CODE (ERROR_DLGACQUIRE_UNKNOWN_FILEDIALOG_SIZE))
   }

   setWindowTitle (tr ("Acquisition parameters", "Dialog title"));
   pOwn    = new t_DlgAcquireLocal;
   pOwn->pDevice = pDevice;

   pLayout = new QVBoxLayout (this);
   setLayout (pLayout);

   pLayout->addWidget (new QLabel(("<b>" + tr("Acquisition parameters for %1") + "</b>") .arg(pDevice->LinuxDevice) , this));

   // Format box (with EWF fields)
   // ----------------------------
   QGroupBox   *pGroupBoxFormat = new QGroupBox (tr("File format"), this);
   QVBoxLayout *pLayoutFormat   = new QVBoxLayout;
   QGridLayout *pLayoutEwf      = new QGridLayout ();

   pGroupBoxFormat->setLayout (pLayoutFormat);
   pLayout->addWidget (pGroupBoxFormat);
   pLayout->addStretch (1);

   CHK_EXIT (t_File::GetFormatDescription (t_File::EWF, Str))            ButtonTextEWF = "&" + Str;
   CHK_EXIT (t_File::GetFormatDescription (t_File::DD , Str))            ButtonTextDD  = "&" + Str;
   CHK_EXIT (t_File::GetFormatExtension   (t_File::EWF, NULL, &Str))     ButtonTextEWF+= " " + tr("(file extension %1)") .arg (Str);
   CHK_EXIT (t_File::GetFormatExtension   (t_File::DD , NULL, &Str))     ButtonTextDD += " " + tr("(file extension %1)") .arg (Str);

   pOwn->pRadioButtonFormatEWF = new QRadioButton (ButtonTextEWF, this);
   pOwn->pRadioButtonFormatDD  = new QRadioButton (ButtonTextDD , this);

   pLayoutFormat->addWidget  (pOwn->pRadioButtonFormatDD );
   pLayoutFormat->addWidget  (pOwn->pRadioButtonFormatEWF);
   pLayoutFormat->addLayout  (pLayoutEwf);

   CHK_EXIT (CfgGetDlgAcquireFields (&pDlgAcquireFields))

   for (i=0; i<pDlgAcquireFields->count(); i++)
   {
      pDlgAcquireField = pDlgAcquireFields->at(i);
      if (pDlgAcquireField->FieldName.startsWith (CFG_DLGACQUIRE_EWF_FIELDID, Qt::CaseInsensitive))
      {
         pLabel    = new QLabel (tr(QSTR_TO_PSZ(pDlgAcquireField->FieldName)), this);
         pLineEdit = new t_DlgAcquireLineEdit (this, pDlgAcquireField->FieldName);
         pDlgAcquireField->pLineEdit = pLineEdit;
         if (pDlgAcquireField->EntryMode == CFG_ENTRYMODE_HIDE)
         {
            pLabel   ->hide();
            pLineEdit->hide();
         }
         else
         {
            pLayoutEwf->addWidget (pLabel   , Row, 1);
            pLayoutEwf->addWidget (pLineEdit, Row, 2);
            pOwn->EwfWidgetsList.append (pLabel   );
            pOwn->EwfWidgetsList.append (pLineEdit);
            Row++;
         }
         CHK_EXIT (DlgAcquireGetFieldValue (pDevice, pDlgAcquireField, Set))
         pLineEdit->setText (Set);
         CHK_QT_EXIT (connect (pLineEdit, SIGNAL (SignalTextEdited (t_DlgAcquireLineEdit *, const QString &)),
                                    this, SLOT   (SlotTextEdited   (t_DlgAcquireLineEdit *, const QString &))))
      }
   }

   pLayoutEwf->setColumnMinimumWidth (0, 20);

   // Destination box
   // ---------------
   QGroupBox   *pGroupBoxDest = new QGroupBox(tr("Destination"), this);
   QGridLayout *pLayoutDest   = new QGridLayout ();
   pGroupBoxDest->setLayout (pLayoutDest);
   pLayout->addWidget (pGroupBoxDest);
   pLayout->addStretch (1);

   Row = 0;
   if (CONFIG(WriteToDevNull))
   {
      pLabel = new QLabel(tr("Configuration flag WriteToDevNull is set!"), this);
      pLayoutDest->addWidget (pLabel, Row++, 0);
   }
   //lint -restore

   for (i=0; i<pDlgAcquireFields->count(); i++)
   {
      pDlgAcquireField = pDlgAcquireFields->at(i);
      if (pDlgAcquireField->FieldName.startsWith (CFG_DLGACQUIRE_DEST_FIELDID, Qt::CaseInsensitive))
      {
         DestDirectory = (pDlgAcquireField->FieldName.compare(CFG_DLGACQUIRE_DEST_DIRECTORY, Qt::CaseInsensitive) == 0);
         pLabel = new QLabel (tr(QSTR_TO_PSZ(pDlgAcquireField->FieldName)), this);

         if (DestDirectory)
         {
            pLayoutDest->addWidget (pLabel, Row, 0);              // Use 1st column for label
            pButtonBrowse = new QPushButton (tr("..."), this);    // and 2nd for browse button
            pLayoutDest->addWidget (pButtonBrowse, Row, 1);
            pButtonBrowse->setSizePolicy (QSizePolicy(QSizePolicy::Ignored, QSizePolicy::Preferred));   // Allow to shrink horizontal size below preferred minimum
         }
         else
         {
            pLayoutDest->addWidget (pLabel, Row, 0, 1, 2);        // Use the first 2 column for label
         }
         pLineEdit = new t_DlgAcquireLineEdit (this, pDlgAcquireField->FieldName);
         pDlgAcquireField->pLineEdit = pLineEdit;
         pLayoutDest->addWidget (pLineEdit, Row, 2);

         CHK_EXIT (DlgAcquireGetFieldValue (pDevice, pDlgAcquireField, Set))
         if (DestDirectory)
         {
            pLineEdit->setReadOnly (true);
            if (!Set.endsWith ("/"))
               Set += "/";
         }
         pLineEdit->setText (Set);

         CHK_QT_EXIT (connect (pLineEdit, SIGNAL (SignalTextEdited (t_DlgAcquireLineEdit *, const QString &)),
                                    this, SLOT   (SlotTextEdited   (t_DlgAcquireLineEdit *, const QString &))))
         Row++;
      }
   }
   pOwn->pLineEditDestDirectory     = DlgAcquireGetLineEdit (CFG_DLGACQUIRE_DEST_DIRECTORY    );
   pOwn->pLineEditDestImageFilename = DlgAcquireGetLineEdit (CFG_DLGACQUIRE_DEST_IMAGEFILENAME);
   pOwn->pLineEditDestInfoFilename  = DlgAcquireGetLineEdit (CFG_DLGACQUIRE_DEST_INFOFILENAME );

   // Hash
   // ----
   QGroupBox   *pGroupBoxHash = new QGroupBox(tr("Hash computation"), this);
   QGridLayout *pLayoutHash   = new QGridLayout ();
   pGroupBoxHash->setLayout (pLayoutHash);
   pLayout->addWidget (pGroupBoxHash);
   pLayout->addStretch (1);

   pOwn->pCheckBoxCalcMD5   = new QCheckBox (tr("Calculate MD5 hash"), this);
   pOwn->pCheckBoxVerifyMD5 = new QCheckBox (tr("Verify MD5 hash (not implemented yet)"   ), this);
   pLayoutHash->addWidget (pOwn->pCheckBoxCalcMD5  , 0, 0);
   pLayoutHash->addWidget (pOwn->pCheckBoxVerifyMD5, 1, 0);

   // Dialog buttons
   // --------------
   QHBoxLayout *pLayoutButtons = new QHBoxLayout ();
   pLayout->addLayout (pLayoutButtons);

   pOwn->pButtonOk     = new QPushButton (QObject::tr("Ok"    ), this);
   pOwn->pButtonCancel = new QPushButton (QObject::tr("Cancel"), this);
   pLayoutButtons->addWidget (pOwn->pButtonOk    );
   pLayoutButtons->addWidget (pOwn->pButtonCancel);
   pOwn->pButtonOk->setDefault (true);

   // Set default values
   // ------------------
   pOwn->pRadioButtonFormatEWF->setChecked (true);
   pOwn->pCheckBoxCalcMD5     ->setChecked (true);

   UpdateCheckboxState (pOwn->pCheckBoxCalcMD5->isChecked() ? Qt::Checked : Qt::Unchecked);
   UpdateButtonState   ();

   // Connections
   // -----------
   CHK_QT_EXIT (connect (pOwn->pCheckBoxCalcMD5, SIGNAL (stateChanged(int)), this, SLOT(UpdateCheckboxState(int))))

   CHK_QT_EXIT (connect (pOwn->pLineEditDestImageFilename, SIGNAL (textChanged(const QString &)), this, SLOT(UpdateButtonState(const QString &))))
   CHK_QT_EXIT (connect (pOwn->pLineEditDestInfoFilename , SIGNAL (textChanged(const QString &)), this, SLOT(UpdateButtonState(const QString &))))

   CHK_QT_EXIT (connect (pOwn->pRadioButtonFormatDD , SIGNAL (released()), this, SLOT(UpdateFieldState())))
   CHK_QT_EXIT (connect (pOwn->pRadioButtonFormatEWF, SIGNAL (released()), this, SLOT(UpdateFieldState())))

   CHK_QT_EXIT (connect (pButtonBrowse      , SIGNAL (released()), this, SLOT(SlotBrowse())))
   CHK_QT_EXIT (connect (pOwn->pButtonOk    , SIGNAL (released()), this, SLOT(SlotAccept())))
   CHK_QT_EXIT (connect (pOwn->pButtonCancel, SIGNAL (released()), this, SLOT(reject    ())))
}

void t_DlgAcquire::UpdateCheckboxState (int State)
{
   pOwn->pCheckBoxVerifyMD5->setEnabled (State == Qt::Checked);
}

void t_DlgAcquire::UpdateButtonState (const QString & /*NewText*/)
{
   bool Enabled;

   Enabled = !pOwn->pLineEditDestDirectory    ->text().isEmpty() &&
             !pOwn->pLineEditDestImageFilename->text().isEmpty() &&
             !pOwn->pLineEditDestInfoFilename ->text().isEmpty();
   pOwn->pButtonOk->setEnabled (Enabled);
}

void t_DlgAcquire::UpdateFieldState (void)
{
   QWidget *pWidget;
   bool      Enabled;
   int       i;

   Enabled = pOwn->pRadioButtonFormatEWF->isChecked();

   for (i = 0;
        i < pOwn->EwfWidgetsList.count();
        i++)
   {
      pWidget = pOwn->EwfWidgetsList.at(i);
      pWidget->setEnabled (Enabled);
   }
}

void t_DlgAcquire::SlotBrowse (void)
{
   QString Path;
   bool    OkPressed;

   if (CONFIG(UseFileDialogFromQt))
   {
      Path = QFileDialog::getExistingDirectory(this, tr("Select destination directory", "Dialog title"), pOwn->pLineEditDestDirectory->text(), QFileDialog::ShowDirsOnly);
      OkPressed = !Path.isNull();
   }
   else
   {
      Path = pOwn->pLineEditDestDirectory->text();
      t_DlgDirSel Dlg (&Path, this);

      switch (CONFIG(FileDialogSize))
      {
         case CFG_STARTUPSIZE_STANDARD  :                       break;
         case CFG_STARTUPSIZE_FULLSCREEN: Dlg.showFullScreen(); break;
         case CFG_STARTUPSIZE_MAXIMIZED : Dlg.showMaximized (); break;
         case CFG_STARTUPSIZE_MANUAL    : CHK_EXIT (QtUtilSetGeometryCentered (&Dlg, CONFIG(FileDialogSizeManualDx), CONFIG(FileDialogSizeManualDy))); break;
         default                        : CHK_EXIT (ERROR_DLGACQUIRE_UNKNOWN_FILEDIALOG_SIZE)
      }
      OkPressed = (Dlg.exec() == QDialog::Accepted);
   }

   if (OkPressed)
   {
      if (!Path.endsWith ("/"))
         Path += "/";
      pOwn->pLineEditDestDirectory->setText(Path);
   }
}

void t_DlgAcquire::SlotTextEdited (t_pDlgAcquireLineEdit pLineEdit, const QString &/*NewVal*/)
{
   t_pCfgDlgAcquireRules pDlgAcquireRules;
   t_pCfgDlgAcquireRule  pDlgAcquireRule;
   t_pDlgAcquireLineEdit pLineEditDest;
   int                    i;
   QString                Set;

//   printf ("\nNewVal in %s: %s", QSTR_TO_PSZ(pLineEdit->Name), QSTR_TO_PSZ(NewVal));

   CHK_EXIT (CfgGetDlgAcquireRules (&pDlgAcquireRules))

   for (i=0; i<pDlgAcquireRules->count(); i++)
   {
      pDlgAcquireRule = pDlgAcquireRules->at(i);
      if (pDlgAcquireRule->TriggerFieldName.compare (pLineEdit->Name, Qt::CaseInsensitive) == 0)
      {
         pLineEditDest = DlgAcquireGetLineEdit (pDlgAcquireRule->DestFieldName);

         CHK_EXIT (DlgAcquireResolveSpecialSequences (pOwn->pDevice, true, pDlgAcquireRule->Value, Set))
         pLineEditDest->setText(Set);
      }
   }
}

static APIRET DlgAcquireCheckFilename (const QString &In, bool &Clean, QString &Out)
{
   unsigned char Ch;
   int           i;
   bool          Ok;
   QString       SpecialChars = CONFIG (SpecialFilenameChars);

   Clean = true;
   Out   = "";
   for (i=0; i<In.length(); i++)
   {
      Ch = In[i].toAscii();
      Ok = ((Ch >= '0') && (Ch <= '9')) ||
           ((Ch >= 'a') && (Ch <= 'z')) ||
           ((Ch >= 'A') && (Ch <= 'Z')) ||
            (Ch == '_') || SpecialChars.contains(Ch);
      if (Ok)
           Out += Ch;
      else Clean = false;
   }
   if (Out.isEmpty())
      Out = DLG_ACQUIRE_DEFAULT_EMPTY_FILENAME;

   return NO_ERROR;
}

void t_DlgAcquire::SlotAccept (void)
{
   QMessageBox::StandardButton  Button;
   QString                      ImageFilenameCorrected;
   QString                      InfoFilenameCorrected;
   bool                         ImageFilenameClean;
   bool                         InfoFilenameClean;

   CHK_EXIT (DlgAcquireCheckFilename (pOwn->pLineEditDestImageFilename->text(), ImageFilenameClean, ImageFilenameCorrected))
   CHK_EXIT (DlgAcquireCheckFilename (pOwn->pLineEditDestInfoFilename ->text(), InfoFilenameClean,  InfoFilenameCorrected ))

   if (ImageFilenameClean && InfoFilenameClean)
   {
      accept();
   }
   else
   {
      LOG_INFO ("Unallowed characters in filenames found")
      Button = QMessageBox::question (this, tr ("Special characters", "Dialog title"),
                                            tr ("The filenames contain special characters which are not allowed. Guymager suggests the "
                                                "following changes:"
                                                "\n\tImage filename: %1"
                                                "\n\tInfo filename: %2"
                                                "\nDo you accept these changes?") .arg(ImageFilenameCorrected) .arg(InfoFilenameCorrected),
                                                QMessageBox::Yes | QMessageBox::No, QMessageBox::Yes);
      if (Button == QMessageBox::Yes)
      {
         LOG_INFO ("User accepts filename changes")
         pOwn->pLineEditDestImageFilename->setText(ImageFilenameCorrected);
         pOwn->pLineEditDestInfoFilename ->setText(InfoFilenameCorrected );
         accept();
      }
      else
      {
         LOG_INFO ("User rejects filename changes")
      }
   }
}


APIRET t_DlgAcquire::GetParameters (t_Device::t_Acquisition &Acquisition)
{
   t_pCfgDlgAcquireField pDlgAcquireField;

   Acquisition.CalcMD5       = pOwn->pCheckBoxCalcMD5     ->isChecked();
   Acquisition.VerifyMD5     = pOwn->pCheckBoxVerifyMD5   ->isChecked();
   Acquisition.Format        = pOwn->pRadioButtonFormatEWF->isChecked() ? t_File::EWF : t_File::DD;

   #define COPY_VALUE(Acq, FieldName)                    \
      pDlgAcquireField = DlgAcquireGetField (FieldName); \
      Acq = pDlgAcquireField->pLineEdit->text();         \
      pDlgAcquireField->LastEnteredValue = Acq;

   COPY_VALUE (Acquisition.CaseNumber    , CFG_DLGACQUIRE_EWF_CASENUMBER    )
   COPY_VALUE (Acquisition.EvidenceNumber, CFG_DLGACQUIRE_EWF_EVIDENCENUMBER)
   COPY_VALUE (Acquisition.Examiner      , CFG_DLGACQUIRE_EWF_EXAMINER      )
   COPY_VALUE (Acquisition.Description   , CFG_DLGACQUIRE_EWF_DESCRIPTION   )
   COPY_VALUE (Acquisition.Notes         , CFG_DLGACQUIRE_EWF_NOTES         )
   COPY_VALUE (Acquisition.Path          , CFG_DLGACQUIRE_DEST_DIRECTORY    )
   COPY_VALUE (Acquisition.ImageFilename , CFG_DLGACQUIRE_DEST_IMAGEFILENAME)
   COPY_VALUE (Acquisition.InfoFilename  , CFG_DLGACQUIRE_DEST_INFOFILENAME )
   #undef COPY_VALUE

   return NO_ERROR;
}

t_DlgAcquire::~t_DlgAcquire ()
{
   delete pOwn;
}

