// ****************************************************************************
//  Project:        GUYMAGER
// ****************************************************************************
//  Programmer:     Guy Voncken
//                  Police Grand-Ducale
//                  Service de Police Judiciaire
//                  Section Nouvelles Technologies
// ****************************************************************************
//  Module:         The table widget (central widget of the application)
// ****************************************************************************

#include <QtGui>

#include "toolconstants.h"
#include "common.h"
#include "config.h"
#include "threadread.h"
#include "threadwrite.h"
#include "threadmd5.h"
#include "threadcompress.h"
#include "dlgmessage.h"
#include "dlgacquire.h"
#include "itemdelegate.h"
#include "table.h"
#include "compileinfo.h"
#include "dlgabort.h"


class t_TableLocal
{
   public:
      t_pDeviceList  pDeviceList;
      QAction       *pActionAcquire;
      QAction       *pActionAbort;
      QAction       *pActionDeviceInfo;
      bool            SlowDownAcquisitions;
};

t_Table::t_Table ()
{
   CHK_EXIT (ERROR_TABLE_CONSTRUCTOR_NOT_SUPPORTED)
} //lint !e1401 not initialised

t_Table::t_Table (QWidget *pParent, t_pDeviceList pDeviceList)
   :QTableView (pParent)
{
   CHK_EXIT (TOOL_ERROR_REGISTER_CODE (ERROR_TABLE_INVALID_ACTION                ))
   CHK_EXIT (TOOL_ERROR_REGISTER_CODE (ERROR_TABLE_THREADREAD_ALREADY_RUNNING    ))
   CHK_EXIT (TOOL_ERROR_REGISTER_CODE (ERROR_TABLE_THREADMD5_ALREADY_RUNNING     ))
   CHK_EXIT (TOOL_ERROR_REGISTER_CODE (ERROR_TABLE_THREADWRITE_ALREADY_RUNNING   ))
   CHK_EXIT (TOOL_ERROR_REGISTER_CODE (ERROR_TABLE_THREADCOMPRESS_ALREADY_RUNNING))
   CHK_EXIT (TOOL_ERROR_REGISTER_CODE (ERROR_TABLE_THREADREAD_DOESNT_EXIT        ))
   CHK_EXIT (TOOL_ERROR_REGISTER_CODE (ERROR_TABLE_FIFO_EXISTS                   ))
   CHK_EXIT (TOOL_ERROR_REGISTER_CODE (ERROR_TABLE_CONSTRUCTOR_NOT_SUPPORTED     ))

   setItemDelegate      (new t_ItemDelegate (this));
   setSelectionBehavior (QAbstractItemView::SelectRows     );
   setSelectionMode     (QAbstractItemView::SingleSelection);

   pOwn = new t_TableLocal;
   pOwn->pDeviceList          = pDeviceList;
   pOwn->SlowDownAcquisitions = false;

   verticalHeader  ()->hide                  ();
   horizontalHeader()->setClickable          (true);
   horizontalHeader()->setSortIndicatorShown (true);

   pOwn->pActionAcquire    = new QAction(tr("Acquire"          , "Context menu"), this);  addAction (pOwn->pActionAcquire   );
   pOwn->pActionAbort      = new QAction(tr("Abort acquisition", "Context menu"), this);  addAction (pOwn->pActionAbort     );
   pOwn->pActionDeviceInfo = new QAction(tr("Info"             , "Context menu"), this);  addAction (pOwn->pActionDeviceInfo);

   CHK_QT_EXIT (connect (this, SIGNAL(clicked (const QModelIndex &)), this, SLOT(SlotMouseClick(const QModelIndex &))))

//   pOwn->pTable->setContextMenuPolicy (Qt::ActionsContextMenu);
}

t_Table::~t_Table ()
{
   delete pOwn;
}

APIRET t_Table::GetDeviceUnderCursor (t_pDevice &pDevice)
{
   QModelIndex Index;
   int         Row;
   int         Count;

   pDevice = NULL;
   QModelIndexList IndexList = selectionModel()->selectedRows();
   Count = IndexList.count();
   switch (Count)
   {
      case 0:  break;
      case 1:  Index = IndexList.first();
               Row = model()->data(Index, t_ItemDelegate::RowNrRole).toInt();
               pDevice = pOwn->pDeviceList->at(Row);
               break;
      default: LOG_ERROR ("Strange, several rows seem to be selected.")
   }

   return NO_ERROR;
}

void t_Table::contextMenuEvent (QContextMenuEvent *pEvent)
{
   QAction     *pAction;
   QAction       LocalIndicator (tr("Local Device - cannot be acquired"), this);
   QMenu         Menu;
   QPoint        Pos;
   t_pDevice    pDevice;
   int           CorrectedY;
   int           Row;
   QModelIndex   Index;
   bool          Running;

   Pos = mapFromGlobal (pEvent->globalPos());           // The event coordinates are absolute screen coordinates, map them withb this widget
   CorrectedY = Pos.y() - horizontalHeader()->height(); // QTableView::rowAt seems to have a bug: It doesn't include the header height in the calculations. So, we correct it here.
   Pos.setY (CorrectedY);

   Index = indexAt (Pos);
   selectRow (Index.row());
   Row = model()->data(Index, t_ItemDelegate::RowNrRole).toInt();

   if ((Row < 0) || (Row >= pOwn->pDeviceList->count())) // User clicked in the empty area
      return;

   pDevice = pOwn->pDeviceList->at(Row);
//   LOG_INFO ("Context global %d -- widget %d -- corrected %d -- index %d", pEvent->globalPos().y(), Pos.y(), CorrectedY, RowIndex);

   Running = pDevice->pThreadRead || pDevice->pThreadWrite;

   pOwn->pActionAcquire->setEnabled (!Running);
   pOwn->pActionAbort  ->setEnabled ( Running);

   if (pDevice->Local)
   {
      LocalIndicator.setEnabled (false);
      Menu.addAction (&LocalIndicator);
      Menu.addSeparator ();
   }
   else
   {
      Menu.addAction (pOwn->pActionAcquire);
      Menu.addAction (pOwn->pActionAbort  );
   }
   Menu.addAction (pOwn->pActionDeviceInfo);

   pAction = Menu.exec (pEvent->globalPos());

//   pAction = QMenu::exec (actions(), pEvent->globalPos());
   if      (pAction == pOwn->pActionAcquire   ) CHK_EXIT (StartAcquisition  (pDevice))
   if      (pAction == pOwn->pActionAbort     ) CHK_EXIT (AbortAcquisition  (pDevice))
   else if (pAction == pOwn->pActionDeviceInfo) CHK_EXIT (ShowDeviceInfo    (pDevice))
   else
   {
//      CHK_EXIT (ERROR_TABLE_INVALID_ACTION)
   }
}

void t_Table::SlotMouseClick (const QModelIndex & Index)
{
   int Row;

   selectRow (Index.row());
   Row = model()->data (Index, t_ItemDelegate::RowNrRole).toInt();
   emit (SignalDeviceSelected (pOwn->pDeviceList->at(Row)));
}


APIRET t_Table::ShowDeviceInfo (t_pcDevice pDevice)
{
   QString Info;

   LOG_INFO ("Info for %s requested", QSTR_TO_PSZ(pDevice->LinuxDevice))
   pOwn->SlowDownAcquisitions = true;
   CHK (t_Info::GetDeviceInfo (pDevice, true, Info))
   pOwn->SlowDownAcquisitions = false;
   CHK (t_DlgMessage::Show (tr("Device info", "Dialog title"), Info, true))

   return NO_ERROR;
}

// ----------------------------------------------------------------------------------------------------------------
//
//                                                     Acquistion start
//
// ----------------------------------------------------------------------------------------------------------------



APIRET t_Table::InfoAcquisitionStart (t_pDevice pDevice)
{
   QString FormatDescription;
   QString FormatExtension;

   CHK (pDevice->Info.Create())
   CHK (pDevice->Info.Title   (tr("GUYMAGER ACQUISITION INFO FILE", "Info file")))

   CHK (pDevice->Info.WriteLn ())
   CHK (pDevice->Info.Title   (tr("Guymager", "Info file")))
   CHK (pDevice->Info.WriteLn ())
   CHK (pDevice->Info.WriteLn (tr("Compile timestamp: %1", "Info file") .arg(pCompileInfoTimestamp)))
   CHK (pDevice->Info.WriteLn (tr("Version: %1"          , "Info file") .arg(pCompileInfoVersion)))

   CHK (pDevice->Info.WriteLn ())
   CHK (pDevice->Info.WriteLn ())
   CHK (pDevice->Info.Title   (tr("Device information", "Info file")))
   CHK (pDevice->Info.WriteLn ())
   CHK (pDevice->Info.WriteDeviceInfo ())

   CHK (pDevice->Info.WriteLn ())
   CHK (pDevice->Info.WriteLn ())
   CHK (pDevice->Info.Title   (tr("Image", "Info file")))
   CHK (pDevice->Info.WriteLn ())
   CHK (pDevice->Info.WriteLn (tr("Linux device: %1"    , "Info file").arg(pDevice->LinuxDevice) ))
   CHK (pDevice->Info.WriteLn (tr("Device size: %1"     , "Info file").arg(pDevice->Size)))
   CHK (pDevice->Info.WriteLn (tr("Destination path: %1", "Info file").arg(pDevice->Acquisition.Path)))
   CHK (pDevice->Info.WriteLn (tr("Image file name: %1" , "Info file").arg(pDevice->Acquisition.ImageFilename)))
   CHK (pDevice->Info.WriteLn (tr("Info file name: %1"  , "Info file").arg(pDevice->Acquisition.InfoFilename )))

   CHK (t_File::GetFormatDescription (pDevice->Acquisition.Format, FormatDescription))
   CHK (t_File::GetFormatExtension   (pDevice->Acquisition.Format, NULL, &FormatExtension))

   CHK (pDevice->Info.WriteLn (tr("Image format: %1 - file extension is %2", "Info file")  .arg(FormatDescription) .arg(FormatExtension)))

   return NO_ERROR;
}

APIRET t_Table::StartAcquisition (t_pDevice pDevice)
{
   t_DlgAcquire Dlg (pDevice, this);
   QString      Format;

   if (Dlg.exec() == QDialog::Accepted)
   {
      pOwn->SlowDownAcquisitions = true;
      LOG_INFO ("Starting acquisition for %s", QSTR_TO_PSZ(pDevice->LinuxDevice))
      CHK (Dlg.GetParameters (pDevice->Acquisition))
      CHK (t_File::GetFormatExtension   (pDevice->Acquisition.Format, NULL, &Format))
      LOG_INFO ("Destination folder %s, image %s%s, info %s%s", QSTR_TO_PSZ(pDevice->Acquisition.Path),
                                                                QSTR_TO_PSZ(pDevice->Acquisition.ImageFilename), QSTR_TO_PSZ(Format),
                                                                QSTR_TO_PSZ(pDevice->Acquisition.InfoFilename ), t_File::pExtensionInfo)
      CHK (pDevice->SetMessage (QString()))
      pDevice->AcquireState    = t_Device::Running;
      pDevice->AbortReason     = t_Device::None;
      pDevice->AbortRequest    = false;
      pDevice->DeleteAfterAbort= false;
      pDevice->StopTimestamp   = QDateTime();
      pDevice->PrevTimestamp   = QTime();
      pDevice->PrevSpeed       = 0.0;
      pDevice->PrevWritePos    = 0;
      pDevice->SetCurrentWritePos(0LL);
      CHK (pDevice->ClearBadSectors ())

      memset (pDevice->MD5Digest, 0, sizeof(pDevice->MD5Digest));
      pDevice->StartTimestamp = QDateTime::currentDateTime();
      CHK (InfoAcquisitionStart (pDevice))

      if (pDevice->pThreadRead  != NULL) CHK_EXIT (ERROR_TABLE_THREADREAD_ALREADY_RUNNING)
      if (pDevice->pThreadWrite != NULL) CHK_EXIT (ERROR_TABLE_THREADWRITE_ALREADY_RUNNING)
      if (pDevice->pThreadMD5   != NULL) CHK_EXIT (ERROR_TABLE_THREADMD5_ALREADY_RUNNING)
      if (!pDevice->ThreadCompressList.isEmpty()) CHK_EXIT (ERROR_TABLE_THREADCOMPRESS_ALREADY_RUNNING)
      if ((pDevice->pFifoRead        != NULL) ||
          (pDevice->pFifoMD5In       != NULL) ||
          (pDevice->pFifoMD5Out      != NULL) ||
          (pDevice->pFifoCompressIn  != NULL) ||
          (pDevice->pFifoCompressOut != NULL))
         CHK_EXIT (ERROR_TABLE_FIFO_EXISTS)

      // Create threads
      // --------------
      pDevice->pThreadRead  = new t_ThreadRead  (pDevice, &pOwn->SlowDownAcquisitions);
      pDevice->pThreadWrite = new t_ThreadWrite (pDevice, &pOwn->SlowDownAcquisitions);
      CHK_QT_EXIT (connect (pDevice->pThreadRead , SIGNAL(SignalEnded(t_pDevice)), this, SLOT(SlotThreadReadFinished (t_pDevice))))
      CHK_QT_EXIT (connect (pDevice->pThreadWrite, SIGNAL(SignalEnded(t_pDevice)), this, SLOT(SlotThreadWriteFinished(t_pDevice))))

      if (pDevice->HasHashThread())
      {
         pDevice->pThreadMD5  = new t_ThreadMD5 (pDevice);
         CHK_QT_EXIT (connect (pDevice->pThreadMD5, SIGNAL(SignalEnded(t_pDevice)), this, SLOT(SlotThreadMD5Finished (t_pDevice))))
      }

      if (pDevice->HasCompressionThreads())
      {
         for (int i=0; i < CONFIG(CompressionThreads); i++)
         {
            t_pThreadCompress pThread;

            pThread = new t_ThreadCompress (pDevice, i);
            pDevice->ThreadCompressList.append (pThread);
            CHK_QT_EXIT (connect (pThread, SIGNAL(SignalEnded(t_pDevice,int)), this, SLOT(SlotThreadCompressFinished (t_pDevice,int))))
         }
      }

      // Create fifos and assign threads
      // -------------------------------

      if      (!pDevice->HasHashThread() && !pDevice->HasCompressionThreads())
      {
         pDevice->pFifoRead   = new t_FifoStd (CONFIG (FifoMaxBlocks));
         pDevice->pFifoWrite  = pDevice->pFifoRead;
      }
      else if ( pDevice->HasHashThread() && !pDevice->HasCompressionThreads())
      {
         pDevice->pFifoRead   = new t_FifoStd (CONFIG (FifoMaxBlocks));
         pDevice->pFifoMD5In  = pDevice->pFifoRead;
         pDevice->pFifoMD5Out = new t_FifoStd (CONFIG (FifoMaxBlocks));
         pDevice->pFifoWrite  = pDevice->pFifoMD5Out;
      }
      else if (!pDevice->HasHashThread() &&  pDevice->HasCompressionThreads())
      {
         pDevice->pFifoCompressIn  = new t_FifoCompressIn  (CONFIG(CompressionThreads), CONFIG (FifoMaxBlocks));
         pDevice->pFifoRead        = pDevice->pFifoCompressIn;
         pDevice->pFifoCompressOut = new t_FifoCompressOut (CONFIG(CompressionThreads), CONFIG (FifoMaxBlocks));
         pDevice->pFifoWrite       = pDevice->pFifoCompressOut;
      }
      else if ( pDevice->HasHashThread() &&  pDevice->HasCompressionThreads())
      {
         pDevice->pFifoRead        = new t_FifoStd (CONFIG (FifoMaxBlocks));
         pDevice->pFifoMD5In       = pDevice->pFifoRead;
         pDevice->pFifoCompressIn  = new t_FifoCompressIn  (CONFIG(CompressionThreads), CONFIG (FifoMaxBlocks));
         pDevice->pFifoMD5Out      = pDevice->pFifoCompressIn;
         pDevice->pFifoCompressOut = new t_FifoCompressOut (CONFIG(CompressionThreads), CONFIG (FifoMaxBlocks));
         pDevice->pFifoWrite       = pDevice->pFifoCompressOut;
      }

      // Start threads
      // -------------
      pDevice->pThreadRead ->start(QThread::LowPriority);
      if (pDevice->HasHashThread())
         pDevice->pThreadMD5->start(QThread::LowPriority);
      for (int i=0; i < pDevice->ThreadCompressList.count(); i++)
         pDevice->ThreadCompressList[i]->start(QThread::LowPriority);
      pDevice->pThreadWrite->start(QThread::LowPriority);

      pOwn->SlowDownAcquisitions = false;
   }

   return NO_ERROR;
}


// ----------------------------------------------------------------------------------------------------------------
//
//                                                     Acquistion end
//
// ----------------------------------------------------------------------------------------------------------------

APIRET t_Table::InfoAcquisitionBadSectors (t_pDevice pDevice)
{
   QList<quint64> BadSectors;
   quint64        Count, i;
   quint64        From, To, Next;
   int            LineEntries    =  0;
   const int      MaxLineEntries = 10;
   bool           First          = true;

   CHK (pDevice->GetBadSectors (BadSectors))
   Count = BadSectors.count();
   if (Count)
   {
      LOG_INFO ("Device contains %lld bad sectors", Count)
      CHK (pDevice->Info.WriteLn (tr("Device contains %1 bad sectors. They have been replaced by zeroed sectors. The sector numbers are:", "Info file") .arg(Count)))
      CHK (pDevice->Info.WriteLn ("   "))
      i = 0;
      while (i<Count)
      {
         From = BadSectors.at(i++);
         To   = From;
         while (i < Count)
         {
            Next = BadSectors.at(i);
            if (Next != To+1)
               break;
            To = Next;
            i++;
         }
         if (!First)
         {
            if (LineEntries >= MaxLineEntries)
            {
               CHK (pDevice->Info.Write   (","))
               CHK (pDevice->Info.WriteLn ("   "))
               LineEntries = 0;
            }
            else
            {
               CHK (pDevice->Info.Write (", "))
            }
         }
         First = false;
         if (From == To)
         {
            CHK (pDevice->Info.Write ("%Ld", From))
            LineEntries++;
         }
         else
         {
            CHK (pDevice->Info.Write ("%Ld-%Ld", From, To))
            LineEntries += 2;
         }
      }
   }
   else
   {
      LOG_INFO ("Device contains no bad sectors")
      CHK (pDevice->Info.WriteLn (tr("Device contains no bad sectors.", "Info file")))
   }

   return NO_ERROR;
}

APIRET t_Table::InfoAcquisitionEnd (t_pDevice pDevice)
{
   QString    Hash;
   QString    StateStr;
   int        i;
   quint64    BadSectors;
   QString    StartStr, StopStr;
   int        Hours, Minutes, Seconds;
   double     Speed;

   CHK (InfoAcquisitionBadSectors (pDevice))
   Hash = tr("Disabled", "Info file, means: Hash disabled");
   switch (pDevice->AcquireState)
   {
      case t_Device::Finished:
         StateStr = tr("Finished successfully", "Info file");
         BadSectors = t_Device::GetBadSectorCount(pDevice).toULongLong();
         if (BadSectors)
            StateStr += " " + tr("(with %1 bad sectors)", "Info file, may be appended to 'finished successfully' message") .arg(BadSectors);
         if (pDevice->Acquisition.CalcMD5)
         {
            Hash.clear();
            for (i=0; i<MD5_DIGEST_LENGTH; i++)
               Hash += QString ("%1") .arg ((int)pDevice->MD5Digest[i], 2, 16, QChar('0'));
         }
         break;

      case t_Device::Aborted:
         switch (pDevice->AbortReason)
         {
            case t_Device::UserRequest:          StateStr = tr("Aborted by user"                     , "Info file"); break;
            case t_Device::ThreadWriteFileError: StateStr = tr("Aborted because of image write error", "Info file"); break;
            default:                             StateStr = tr("Aborted, strange reason (%1)"        , "Info file") .arg(pDevice->AbortReason); break;
         }
         if (pDevice->Acquisition.CalcMD5)
            Hash = tr("Calculation aborted", "Info file");
         break;

      default:
         StateStr = tr("Strange state (%1)", "Info file") .arg(pDevice->AcquireState);
   }
   CHK (pDevice->Info.WriteLn (tr("State: %1"   , "Info file") .arg(StateStr)))
   CHK (pDevice->Info.WriteLn (tr("MD5 hash: %1", "Info file") .arg(Hash    )))


   StartStr = pDevice->StartTimestamp.toString (Qt::ISODate); StartStr.replace ("T", " ");
   StopStr  = pDevice->StopTimestamp.toString  (Qt::ISODate); StopStr .replace ("T", " ");
   Seconds  = pDevice->StartTimestamp.secsTo (pDevice->StopTimestamp);
   Speed    = (double) pDevice->Size / ((double)BYTES_PER_MEGABYTE * (double)Seconds);
   Hours    = Seconds / SECONDS_PER_HOUR  ; Seconds -= Hours   * SECONDS_PER_HOUR;
   Minutes  = Seconds / SECONDS_PER_MINUTE; Seconds -= Minutes * SECONDS_PER_MINUTE;

   CHK (pDevice->Info.WriteLn (tr("Started: %1 (ISO format YYYY-MM-DD HH:MM:SS)"     , "Info file").arg(StartStr)))
   CHK (pDevice->Info.WriteLn (tr("Ended  : %1 (%2 hours, %3 minutes and %4 seconds)", "Info file").arg(StopStr ) .arg(Hours) .arg(Minutes) .arg(Seconds)))
   CHK (pDevice->Info.WriteLn (tr("Speed  : %1 MByte/s"                              , "Info file").arg(Speed, 0, 'f', 2)))
   CHK (pDevice->Info.WriteLn ())

   return NO_ERROR;
}

APIRET t_Table::FinaliseThreadStructs (t_pDevice pDevice)
{
   if ((pDevice->pThreadRead  == NULL) &&        // TableCleanupThreadStructs is called upon every thread's end,
       (pDevice->pThreadWrite == NULL) &&        // but its core only is executed after the last thread ended (as
       (pDevice->pThreadMD5   == NULL) &&        // we must not kill vital structures as long as any threads are
       (pDevice->ThreadCompressList.isEmpty()))  // still running).
   {
      if      (!pDevice->HasHashThread() && !pDevice->HasCompressionThreads())
      {
         delete pDevice->pFifoRead;
      }
      else if ( pDevice->HasHashThread() && !pDevice->HasCompressionThreads())
      {
         delete pDevice->pFifoRead;
         delete pDevice->pFifoMD5Out;
      }
      else if (!pDevice->HasHashThread() &&  pDevice->HasCompressionThreads())
      {
         delete pDevice->pFifoCompressIn;
         delete pDevice->pFifoCompressOut;
      }
      else if ( pDevice->HasHashThread() &&  pDevice->HasCompressionThreads())
      {
         delete pDevice->pFifoRead;
         delete pDevice->pFifoCompressIn;
         delete pDevice->pFifoCompressOut;
      }

      pDevice->pFifoRead        = NULL;
      pDevice->pFifoMD5In       = NULL;
      pDevice->pFifoMD5Out      = NULL;
      pDevice->pFifoWrite       = NULL;
      pDevice->pFifoCompressIn  = NULL;
      pDevice->pFifoCompressOut = NULL;

//      LibEwfGetMemStats (&LibEwfAllocs, &LibEwfFrees);
//      LOG_INFO ("LIBEWF mem  statistics: %d allocated - %d freed = %d remaining", LibEwfAllocs, LibEwfFrees, LibEwfAllocs - LibEwfFrees)

      LOG_INFO ("Acquisition of %s: All structures cleaned.", QSTR_TO_PSZ (pDevice->LinuxDevice))

      if (pDevice->AbortRequest)
           pDevice->AcquireState = t_Device::Aborted;
      else pDevice->AcquireState = t_Device::Finished;

      if (!pDevice->DeleteAfterAbort)
         CHK_EXIT (InfoAcquisitionEnd (pDevice))
   }

   return NO_ERROR;
}

void t_Table::SlotThreadReadFinished (t_pDevice pDevice)
{
   LOG_INFO ("Acquisition of %s: Reading thread finished", QSTR_TO_PSZ (pDevice->LinuxDevice))
   pDevice->pThreadRead->deleteLater();
   pDevice->pThreadRead = NULL;

   if (pDevice->AbortRequest)
        CHK_EXIT (WakeWaitingThreads(pDevice))
   else CHK_EXIT (pDevice->pFifoRead->InsertDummy ())
   CHK_EXIT (FinaliseThreadStructs (pDevice))
}

void t_Table::SlotThreadMD5Finished (t_pDevice pDevice)
{
   LOG_INFO ("Acquisition of %s: MD5 thread finished", QSTR_TO_PSZ (pDevice->LinuxDevice))
   pDevice->pThreadMD5->deleteLater();
   pDevice->pThreadMD5 = NULL;

   if (pDevice->AbortRequest)
        CHK_EXIT (WakeWaitingThreads(pDevice))
   else CHK_EXIT (pDevice->pFifoMD5Out->InsertDummy ())
   CHK_EXIT (FinaliseThreadStructs (pDevice))
}

void t_Table::SlotThreadCompressFinished (t_pDevice pDevice, int ThreadNr)
{
   bool NoMoreThreads = true;

   LOG_INFO ("Acquisition of %s: Compression thread #%d finished", QSTR_TO_PSZ (pDevice->LinuxDevice), ThreadNr)

   pDevice->ThreadCompressList[ThreadNr]->deleteLater();
   pDevice->ThreadCompressList[ThreadNr] = NULL;

   for (int i=0;
        (i < pDevice->ThreadCompressList.count()) && NoMoreThreads;
        i++)
      NoMoreThreads = (pDevice->ThreadCompressList[i] == NULL);

   if (NoMoreThreads)
   {
       LOG_INFO ("All %d compression threads finished", pDevice->ThreadCompressList.count())
       pDevice->ThreadCompressList.clear();
   }

   if (pDevice->AbortRequest)
        CHK_EXIT (WakeWaitingThreads(pDevice))
   else CHK_EXIT (pDevice->pFifoCompressOut->InsertDummy (ThreadNr))

   CHK_EXIT (FinaliseThreadStructs (pDevice))
}

void t_Table::SlotThreadWriteFinished (t_pDevice pDevice)
{
   LOG_INFO ("Acquisition of %s: Writing thread finished", QSTR_TO_PSZ (pDevice->LinuxDevice))

   pDevice->pThreadWrite->deleteLater();
   pDevice->pThreadWrite = NULL;

   if (pDevice->AbortRequest)
      CHK_EXIT (WakeWaitingThreads(pDevice))

   CHK_EXIT (FinaliseThreadStructs (pDevice))
}

APIRET t_Table::AbortAcquisition0 (t_pDevice pDevice)
{
   if (pDevice->pThreadRead == NULL)
   {
      LOG_INFO ("User pressed abort, but acquisition on %s is no longer running (it probably ended between the user's click and this message)", QSTR_TO_PSZ(pDevice->LinuxDevice))
      return NO_ERROR;
   }

   LOG_INFO ("User aborts acquisition on %s", QSTR_TO_PSZ(pDevice->LinuxDevice))
   pDevice->AbortReason  = t_Device::UserRequest;
   pDevice->AbortRequest = true;

   CHK (WakeWaitingThreads (pDevice))

   return NO_ERROR;
}

APIRET t_Table::AbortAcquisition (t_pDevice pDevice)
{
   bool Abort;

   CHK (t_DlgAbort::Show (pDevice, Abort, pDevice->DeleteAfterAbort))
   if (Abort)
      CHK (AbortAcquisition0 (pDevice))

   return NO_ERROR;
}

APIRET t_Table::AbortAllAcquisitions (void)
{
   t_pDevice pDevice;
   int        i;

   for (i=0; i<pOwn->pDeviceList->count(); i++)
   {
      pDevice = pOwn->pDeviceList->at(i);
      if ((pDevice->AcquireState == t_Device::Running) ||
          (pDevice->AcquireState == t_Device::Paused))
      {
         CHK (AbortAcquisition0 (pDevice))
      }
   }

   return NO_ERROR;
}

APIRET t_Table::WakeWaitingThreads (t_pDevice pDevice)
{
   if      (!pDevice->HasHashThread() && !pDevice->HasCompressionThreads())
   {
      CHK (pDevice->pFifoRead->WakeWaitingThreads())
   }
   else if ( pDevice->HasHashThread() && !pDevice->HasCompressionThreads())
   {
      CHK (pDevice->pFifoRead  ->WakeWaitingThreads())
      CHK (pDevice->pFifoMD5Out->WakeWaitingThreads())
   }
   else if (!pDevice->HasHashThread() &&  pDevice->HasCompressionThreads())
   {
      CHK (pDevice->pFifoCompressIn ->WakeWaitingThreads())
      CHK (pDevice->pFifoCompressOut->WakeWaitingThreads())
   }
   else if ( pDevice->HasHashThread() &&  pDevice->HasCompressionThreads())
   {
      CHK (pDevice->pFifoRead       ->WakeWaitingThreads())
      CHK (pDevice->pFifoCompressIn ->WakeWaitingThreads())
      CHK (pDevice->pFifoCompressOut->WakeWaitingThreads())
   }

   return NO_ERROR;
}

